#
# CanDo - Web based competency tracking system.
# Copyright (C) 2006 Arlington Public Schools, Arlington, VA and The
# Shuttleworth Foundation.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Virginia Gradebook views.

$Id:$
"""

from zope.interface import implements
from zope.component import queryMultiAdapter, getMultiAdapter
from zope.publisher.browser import BrowserView
from zope.security.proxy import removeSecurityProxy
from zope.publisher.interfaces import IPublishTraverse
from zope.location.location import LocationProxy
from zope.viewlet.viewlet import CSSViewlet
from zope.viewlet.viewlet import JavaScriptViewlet
from zope.traversing.browser import absoluteURL
from zope.app.renderer.plaintext import PlainTextToHTMLRenderer

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.person.interfaces import IPerson
from schooltool.course.interfaces import ISection

from cando.virginia.competency import getAdditionalCompetencies
from cando.virginia.competency import getCoreCompetencies

from cando.skilldriver.interfaces import ISkillDriver
from cando.skilldriver.interfaces import ISectionSkillDrivers

from cando.iep.interfaces import IStudentIEP

from cando.gradebook import interfaces
from cando.gradebook import gradebook
from cando.gradebook.browser import report
from cando.gradebook.gradebook import _getEvaluation


gradebookStylesheetViewlet = CSSViewlet('gradebook.css')
sectionGradebookStylesheetViewlet = CSSViewlet('section-gradebook.css')
#gradebookJavaScriptViewlet = JavaScriptViewlet('gradebook.js')
highlightFadePluginViewlet = JavaScriptViewlet('jquery.highlightFade.js')


def _breakJSString(origstr):
    newstr = unicode(origstr)
    newstr = u"\\'".join(newstr.split(u"'"))
    newstr = u'\\"'.join(newstr.split(u'"'))
    # XXX: Need a complete list of what needs to be broken
    return newstr


class StudentGradebookTraverser(object):
    '''Traverser that goes from a sections competency gradebook to a
    student competency gradebook using the students username as the
    path in the url.'''

    implements(IPublishTraverse)

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def publishTraverse(self, request, name):
        app = ISchoolToolApplication(None)
        context = removeSecurityProxy(self.context)

        try:
            student = app['persons'][name]
        except KeyError:
            return queryMultiAdapter((self.context, request), name=name)

        try:
            gb = getMultiAdapter((student, context), interfaces.IStudentCompetencyGradebook)
        except ValueError:
            return queryMultiAdapter((self.context, request), name=name)
        # location looks like http://host/path/to/CompetencyGradebook/studentsUsername
        gb = LocationProxy(gb, self.context, name)
        gb.__setattr__('__parent__', gb.__parent__) # XXX THIS IS REALLY BAD XXX
        return gb


class GradebookView(BrowserView):
    def __init__(self, context, request):
        super(GradebookView, self).__init__(context, request)
        self.person = IPerson(self.request.principal)
        self.scoresystem = ISchoolToolApplication(None)['scoresystem']
        self.additionalComps = {}
        self.getSectionAdditionalComps()
        self.getCoursesAdditionalComps()
        self.setFormValues()
        if not self.context.comprefDict:
            newurl = absoluteURL(self.context, self.request) + '/noComps.html'
            self.request.response.redirect(newurl)
            self.hasDict=False
            return
        self.hasDict=True

    def getSectionAdditionalComps(self):
        section = self.getSection()
        for comp in getAdditionalCompetencies(section):
            self.additionalComps[comp] = None

    def getCoursesAdditionalComps(self):
        courses = self.getCourses()
        for course in courses:
            for comp in getAdditionalCompetencies(course):
                self.additionalComps[comp] = None

    def getCourses(self):
        return list(self.getSection().courses)

    def getSection(self):
        section = self.context.__parent__
        if ISkillDriver.providedBy(self.context.__parent__):
            section = self.context.__parent__.__parent__.__parent__
        return removeSecurityProxy(section)

    def isAdditionalCompetency(self, comp):
        return comp in self.additionalComps

    def getCompIdClass(self, comp):
        if self.isAdditionalCompetency(comp):
            return "additional-comp"
        return "%s-comp" % comp.optionalFlag

    def groups(self):
        if ISkillDriver.providedBy(self.context.__parent__):
            skilldrivers = ISectionSkillDrivers(self.getSection()).values()
            return sorted([sd.title for sd in skilldrivers])
        core_groups = []
        additional_groups = []
        for group in self.context.comprefDict:
            if self.context.comprefDict[group]:
                firstComp = self.context.comprefDict[group][0]
                if self.isAdditionalCompetency(firstComp):
                    additional_groups.append(group)
                else:
                    core_groups.append(group)
        return sorted(core_groups, key=self.sortKey) + \
               sorted(additional_groups, key=self.sortKey)

    def sortKey(self, group_title):
        ids = sorted([compref.id
                      for compref in self.context.comprefDict[group_title]])
        if ids:
            return ids[0]
        else:
            return ''

    def time(self):
        from datetime import datetime
        t = datetime.utcnow()
        return "%s-%s-%s %s:%s:%s" % (t.year,t.month,t.day,t.hour,t.minute,t.second)

    def breakJSString(self,oldstr):
        return _breakJSString(oldstr)

    @property
    def students(self):
        return sorted(self.context.students, key=lambda student: student.title)

    @property
    def scores(self):
        return self.scoresystem.scores

    @property
    def comps(self):
        if not self.currentCompGroup:
            return []
        for group in self.context.comprefDict:
            if gradebook.stripCompare(group, self.currentCompGroup):
                comps = [removeSecurityProxy(comp)
                         for comp in self.context.comprefDict[group]]
                break
        else:
            return []
        if ISkillDriver.providedBy(self.context.__parent__):
            core_comps = [comp for comp in comps
                          if not self.isAdditionalCompetency(comp)]
            additional_comps = [comp for comp in comps
                                if self.isAdditionalCompetency(comp)]
            comps = sorted(core_comps, key=lambda x:x.id) + \
                    sorted(additional_comps, key=lambda x:x.id)
        else:
            comps.sort(lambda x,y: cmp(x.id, y.id))
        return comps

    @property
    def reportsAvailable(self):
        return self.context.reports_available

    def getStudentName(self, person):
        return '%s, %s' % (person.last_name, person.first_name)

    def hasIEP(self, person):
        return IStudentIEP(person).hasIEP

    def isGraded(self, evaluation):
        return evaluation is not None and \
               self.scoresystem.getNumericalValue(evaluation.value) is not None

    def setFormValues(self):
        if not self.context.comprefDict:
            return
        self.compGrid = {}
        for student in self.students:
            for comp in self.comps:
                key = '%s_%s' % (student.username, comp.globalId)
                evaluation = self.context.getEvaluation(student, comp)
                if not self.isGraded(evaluation):
                    evalStr = '-'
                else:
                    evalStr = str(evaluation.value.value)
                self.compGrid[key] = evalStr

    def getEvaluation(self, student, comp):
        key = '%s_%s' % (student.username, comp.globalId)
        return self.compGrid[key]

    @property
    def currentCompGroup(self):
        return self.context.getCurrentCompGroup(self.person)

    def isCurrentGroup(self, group):
        return gradebook.stripCompare(group, self.currentCompGroup)

    def update(self):
        self.changed = False
        if 'form-submitted' in self.request and self.request['form-submitted']!='no-edit-content':
            currentCompGroup = self.request.get('selectedCompGroup',
                                                self.currentCompGroup)
            self.context.setCurrentCompGroup(self.person, currentCompGroup)
            for student in self.students:
                for comp in self.comps:
                    key = '%s_%s' % (student.username, comp.globalId)
                    if key not in self.request:
                        continue

                    grade = gradebook.getScoreSystem(comp).fromUnicode(
                        self.request[key])
                    oldgrade = self.context.getEvaluation(student, comp)

                    if oldgrade != grade:
                        self.changed = True
                        self.context.evaluate(student, comp, grade, self.person)
        if 'form-submitted' in self.request:
            if 'selectedCompGroup' in self.request:
                currentCompGroup = self.request['selectedCompGroup']
                if ISkillDriver.providedBy(self.context.__parent__):
                    if not self.changed:
                        drivers = ISectionSkillDrivers(self.getSection()).values()
                        groups = dict([(sd.title, sd) for sd in drivers])
                        if currentCompGroup in groups:
                            newurl = absoluteURL(groups[currentCompGroup],
                                                 self.request) + '/gradeCompetencies'
                            self.request.response.redirect(newurl)
                            return
                else:
                    self.context.setCurrentCompGroup(self.person,
                                                     currentCompGroup)
        self.setFormValues()


class StudentGradebookView(BrowserView):
    def __init__(self, context, request):
        super(StudentGradebookView, self).__init__(context, request)
        self.student = self.context.student
        self.student_iep = IStudentIEP(self.student)
        self.hasIEP = self.student_iep.hasIEP
        self.comment = PlainTextToHTMLRenderer(self.student_iep.comment or '',
                                               self.request).render()
        self.scoresystem = ISchoolToolApplication(None)['scoresystem']
        self.person = IPerson(self.request.principal)
        self.additionalComps = {}
        self.getSectionAdditionalComps()
        self.getCoursesAdditionalComps()
        self.setFormValues()
        if not self.context.comprefDict:
            newurl = absoluteURL(self.context, self.request) + '/noComps.html'
            self.request.response.redirect(newurl)
            self.hasDict=False
            return
        self.hasDict=True
        self.parentSection = self.context
        while not ISection.implementedBy(self.parentSection.__class__):
            self.parentSection = self.parentSection.__parent__
        if self.person in self.parentSection.instructors:
            self.canEdit = True
        else:
            self.canEdit = False

        self.calculateStatistics()

    def getSectionAdditionalComps(self):
        section = self.getSection()
        for comp in getAdditionalCompetencies(section):
            self.additionalComps[comp] = None

    def getCoursesAdditionalComps(self):
        courses = self.getCourses()
        for course in courses:
            for comp in getAdditionalCompetencies(course):
                self.additionalComps[comp] = None

    def getCourses(self):
        return list(self.getSection().courses)

    def getSection(self):
        section = self.context.__parent__.__parent__
        if ISkillDriver.providedBy(self.context.__parent__.__parent__):
            section = self.context.__parent__.__parent__.__parent__.__parent__
        return removeSecurityProxy(section)
        
    def isAdditionalCompetency(self, comp):
        return comp in self.additionalComps

    def sortGroupCompetencies(self, group):
        if ISkillDriver.providedBy(self.context.__parent__.__parent__):
            comps = self.context.comprefDict[group]
            core_comps = [comp for comp in comps
                          if not self.isAdditionalCompetency(comp)]
            additional_comps = [comp for comp in comps
                                if self.isAdditionalCompetency(comp)]
            comps = sorted(core_comps, key=lambda x:x.id) + \
                    sorted(additional_comps, key=lambda x:x.id)
            return comps
        else:
            return sorted(self.context.comprefDict[group], key=lambda x:x.id)

    def groups(self):
        core_groups = []
        additional_groups = []
        for group in self.context.comprefDict:
            if self.context.comprefDict[group]:
                firstComp = self.context.comprefDict[group][0]
                if self.isAdditionalCompetency(firstComp):
                    additional_groups.append(group)
                else:
                    core_groups.append(group)
        return sorted(core_groups, key=self.sortKey) + \
               sorted(additional_groups, key=self.sortKey)

    def sortKey(self, group_title):
        ids = sorted([compref.id
                      for compref in self.context.comprefDict[group_title]])
        if ids:
            return ids[0]
        else:
            return ''

    @property
    def time(self):
        from datetime import datetime
        t = datetime.utcnow()
        return "%s-%s-%s %s:%s:%s" % (t.year,t.month,t.day,t.hour,t.minute,t.second)

    @property
    def comps(self):
        if not self.currentCompGroup:
            return []
        comps = [removeSecurityProxy(comp)
                 for comp in self.context.comprefDict[self.currentCompGroup]]
        if ISkillDriver.providedBy(self.context.__parent__.__parent__):
            core_comps = [comp for comp in comps
                          if not self.isAdditionalCompetency(comp)]
            additional_comps = [comp for comp in comps
                                if self.isAdditionalCompetency(comp)]
            comps = sorted(core_comps, key=lambda x:x.id) + \
                    sorted(additional_comps, key=lambda x:x.id)
        else:
            comps.sort(lambda x,y: cmp(x.id, y.id))
        return comps

    @property
    def currentCompGroup(self):
        return self.context.getCurrentCompGroup(self.person)

    @property
    def reportsAvailable(self):
        return self.context.reports_available

    @property
    def scores(self):
        return self.scoresystem.scores

    def breakJSString(self, origstr):
        return _breakJSString(origstr)

    def getLastEvalDate(self, comp):
        evaluation = self.context.getEvaluation(comp)
        if not self.isGraded(evaluation):
            return "N/A"
        month = ['Jan','Feb','Mar','Apr','May','Jun','Jul','Aug','Sep','Oct','Nov','Dec'][evaluation.time.month-1]
        retstr = "%s %s %s, %s:%s:%s" % (evaluation.time.day, month,
                                         evaluation.time.year,
                                         evaluation.time.hour,
                                         evaluation.time.minute,
                                         evaluation.time.second)
        return retstr

    def getLastEvaluator(self,comp):
        evaluation = self.context.getEvaluation(comp)
        if not self.isGraded(evaluation):
            return ''
        return evaluation.evaluator.title

    def isGraded(self, evaluation):
        return evaluation is not None and \
               self.scoresystem.getNumericalValue(evaluation.value) is not None

    def setFormValues(self):
        if not self.context.comprefDict:
            return
        self.compGrid = {}
        for compGroup in self.context.comprefDict:
            comps = sorted(self.context.comprefDict[compGroup])
            comps.sort(lambda x,y: cmp(x.id, y.id))
            for comp in comps:
                key = '%s_%s_%s' % (self.student.username, compGroup,
                                    comp.globalId)
                evaluation = self.context.getEvaluation(comp)
                if not self.isGraded(evaluation):
                    evalStr = '-'
                else:
                    evalStr = str(evaluation.value.value)
                self.compGrid[key] = evalStr

    def getEvaluation(self, compgrp, comp):
        key = '%s_%s_%s' % (self.student.username, compgrp, comp.globalId)
        return self.compGrid[key]

    def canEdit(self):
        return True

    def isCurrentGroup(self, group):
        return group == self.context.getCurrentCompGroup(self.person)

    def calculateStatistics(self):
        comps = report.getFlatCompetencies(self.context.comprefDict)
        self.all_total = len(comps)
        student_core_required = self.coreRequiredCompetenciesFor(self.student,
                                                                 self.getSection())
        self.required_total = len(student_core_required)
        self.required_evaluated = 0
        self.required_competent = 0
        self.all_evaluated = 0
        self.all_competent = 0

        for comp in comps:
            ev = _getEvaluation(self.student, comp)
            if ev is None or \
               self.scoresystem.getNumericalValue(ev.value) is None:
                continue
            else:
                self.all_evaluated += 1
                if comp in student_core_required:
                    self.required_evaluated += 1
                value = ev.value.value
                # XXX: Hard coded 2 asked by Welsh
                if value >= 2: 
                    self.all_competent += 1
                    if comp in student_core_required:
                        self.required_competent += 1
        try:
            self.required_rating = float(self.required_competent * 100) / \
                                   float(self.required_evaluated)
        except (ZeroDivisionError,):
            self.required_rating = 0.0
        self.required_rating = report.formatPercent(self.required_rating)
        try:
            self.required_completed = float(self.required_competent * 100) / \
                                      float(self.required_total)
        except (ZeroDivisionError,):
            self.required_completed = 0.0
        self.required_completed = report.formatPercent(self.required_completed)
        try:
            self.all_rating = float(self.all_competent * 100) / \
                              float(self.all_evaluated)
        except (ZeroDivisionError,):
            self.all_rating = 0.0
        self.all_rating = report.formatPercent(self.all_rating)

    def getCoreCompetencies(self, section):
        result = {}
        for comp in getCoreCompetencies(section):
            result[comp] = None
        for course in list(section.courses):
            for comp in getCoreCompetencies(course):
                result[comp] = None
        return result.keys()

    def update(self):
        self.changed = False
        if self.canEdit and 'form-submitted' in self.request:
            for compgrp in self.context.comprefDict:
                for comp in self.context.comprefDict[compgrp]:
                    key = '%s_%s_%s' % (self.student.username, compgrp,
                                        comp.globalId)
                    if key not in self.request:
                        continue
                    grade = gradebook.getScoreSystem(comp).fromUnicode(
                        self.request[key])
                    oldgrade = self.context.getEvaluation(comp)
                    if oldgrade != grade:
                        self.changed = True
                        self.context.evaluate(comp, grade, self.person)
        self.setFormValues()
        self.calculateStatistics()

    def isIEPCompetency(self, comp):
        section = self.getSection()
        return self.student_iep.isIEPCompetency(section, comp)

    def getCompIdClass(self, comp):
        if self.isAdditionalCompetency(comp):
            return 'additional-comp'
        if self.isIEPCompetency(comp):
            return 'iep-comp'
        return '%s-comp' % comp.optionalFlag

    def sectionCoreRequiredCompetencies(self, section):
        return [comp for comp in self.getCoreCompetencies(section)
                if comp.optionalFlag == "required"]

    def coreRequiredCompetenciesFor(self, student, section):
        student_iep = IStudentIEP(student)
        section_iep_comps = student_iep.getIEPCompetencies(section)
        if section_iep_comps is None:
            return self.sectionCoreRequiredCompetencies(section)
        return [comp for comp in self.sectionCoreRequiredCompetencies(section)
                if comp.globalId not in section_iep_comps]
