#
# CanDo - Web based competency tracking system.
# Copyright (C) 2006 Arlington Public Schools, Arlington, VA and The
# Shuttleworth Foundation.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

from z3c.menu.simple import menu
from zope.viewlet.manager import WeightOrderedViewletManager
from zope.app.pagetemplate.viewpagetemplatefile import ViewPageTemplateFile
from zope.traversing.browser import absoluteURL
from zope.component import queryUtility
from zope.component.hooks import getSite

from schooltool.course.interfaces import ISection
from cando.newskin import interfaces, PersonDistinguisher
from cando.virginia.interfaces import ICompetencies
from schooltool.term.interfaces import IDateManager


# Classes for global menu items (top-level tabs)
#
class GlobalViewletManager(WeightOrderedViewletManager, PersonDistinguisher):
    """Special viewlet manager that has logic to determine which viewlets
    to render depending on the context."""

    def render(self):
        viewlets = [viewlet for viewlet in self.viewlets
            if viewlet.isAvailable()]
        if not viewlets:
            return ''
        else:
            return u'\n'.join([viewlet.render() for viewlet in viewlets])


class GlobalMenuItem(menu.SimpleMenuItem, PersonDistinguisher):
    """Menu item viewlet generating global/site related links."""

    selectedView = None
    viewURL = u''
    template = ViewPageTemplateFile('templates/top_menu.pt')

    def isAvailable(self):
        return self.isTeacher or self.isStudent

    @property
    def selected(self):
        if (self.selectedView is not None and
            self.selectedView.providedBy(self.__parent__)):
            return True
        return False

    @property
    def url(self):
        siteURL = absoluteURL(getSite(), self.request)
        return siteURL + '/' + self.viewURL


class SectionsItem(GlobalMenuItem):
    title = u'Sections'
    selectedView = interfaces.ISectionView
    weight = 10

    @property
    def viewURL(self):
        term = queryUtility(IDateManager).current_term
        url = absoluteURL(term, self.request) + "/sections/go.html"
        return url[url.find('schoolyears/'):]


class SummaryItem(GlobalMenuItem):
    title = u'Summary'
    viewURL = 'summary.html'
    selectedView = interfaces.ISummaryView
    weight = 20


class SetupItem(GlobalMenuItem):
    title = u'Setup'
    viewURL = 'setup.html'
    selectedView = interfaces.ICompetencyView
    weight = 30

    def isAvailable(self):
        return self.isTeacher


# Classes for context-sensitive menu items (second-level tabs)
#
class SubTabViewletManager(WeightOrderedViewletManager, PersonDistinguisher):
    """Special viewlet manager that has logic to determine which viewlets
    to render depending on the context."""

    def render(self):
        viewlets = [viewlet for viewlet in self.viewlets
            if viewlet.isAvailable()]
        return u'\n'.join([viewlet.render() for viewlet in viewlets])


class SubTabMenuItem(GlobalMenuItem):
    """Menu item viewlet generating context related links."""

    def isAvailable(self):
        return True

    @property
    def url(self):
        contextURL = absoluteURL(self.context, self.request)
        return contextURL + '/' + self.viewURL


class SectionSubTabMenuItem(SubTabMenuItem):
    """Viewlet base for section tab submenu."""

    def isAvailable(self):
        if self.isStudent:
            return False
        context = self.context
        while not ISection.providedBy(context) and context.__parent__:
            if ICompetencies.providedBy(context):
                return False
            context = context.__parent__
        return ISection.providedBy(context)

    @property
    def url(self):
        context = self.context
        while not ISection.providedBy(context) and context.__parent__:
            context = context.__parent__
        contextURL = absoluteURL(context, self.request)
        return contextURL + '/' + self.viewURL


class EvaluationsSubTabMenuItem(SectionSubTabMenuItem):
    title = u'Evaluations'
    viewURL = 'gradeCompetencies'
    weight = 10


class SkillDriversSubTabMenuItem(SectionSubTabMenuItem):
    title = u'Assignments'
    viewURL = 'skilldrivers/@@go.html'
    weight = 20


class MessageBoxesSubTabMenuItem(SectionSubTabMenuItem):
    title = u'Messages'
    viewURL = 'messageboxes/@@go.html'
    weight = 30


class CompetenciesSubTabMenuItem(SectionSubTabMenuItem):
    title = u'Competencies'
    viewURL = 'view_competencies.html'
    weight = 40


class EnrollmentSubTabMenuItem(SectionSubTabMenuItem):
    title = u'Enrollment'
    viewURL = 'learners.html'
    weight = 50


class CompetencySubTabMenuItem(SubTabMenuItem):
    """Viewlet base for competency tab submenu."""

    def isAvailable(self):
        context = self.context
        while not ICompetencies.providedBy(context) and context.__parent__:
            if ISection.providedBy(context):
                return False
            context = context.__parent__
        return ICompetencies.providedBy(context)

    @property
    def url(self):
        siteURL = absoluteURL(getSite(), self.request)
        return siteURL + '/' + self.viewURL

