#
# CanDo - Web based competency tracking system.
# Copyright (C) 2006 Arlington Public Schools, Arlington, VA and The
# Shuttleworth Foundation.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""Skill Driver Implementation

$Id$
"""
__docformat__ = 'restructuredtest'

from decimal import Decimal

import persistent
from zope.interface import implements
import zope.annotation
from zope.component import getUtility
from zope.schema import fieldproperty
from zope.container import contained
from zope.container import btree
from zope.location.location import LocationProxy
from zope.intid.interfaces import IIntIds
from schooltool.course.interfaces import ICourse, ISection
from schooltool.person.interfaces import IPerson
from schooltool.securitypolicy.crowds import Crowd
from schooltool.gradebook.interfaces import IExternalActivity
from schooltool.gradebook.interfaces import IExternalActivities
from schooltool.requirement.scoresystem import UNSCORED

from cando.virginia.interfaces import ICompetencyList
from cando.gradebook.interfaces import ICompetencyGradebook

import interfaces


class SkillDriver(contained.Contained, persistent.Persistent):
    implements(
        interfaces.ISkillDriver,
        zope.annotation.interfaces.IAttributeAnnotatable)

    title = fieldproperty.FieldProperty(interfaces.ISkillDriver['title'])
    description = fieldproperty.FieldProperty(
        interfaces.ISkillDriver['description'])

    def __init__(self, title=u"", description=u""):
        super(SkillDriver, self).__init__()
        self.title = title
        self.description = description

    def getCompetencies(self):
        return ICompetencyList(self)

    def __repr__(self):
        return '<%s %r>' %(self.__class__.__name__, self.title)


class SkillDrivers(btree.BTreeContainer):
    """Implementation of skill driver container."""
    """
    zope.component.adapts(ICourse)
    implements(interfaces.ISkillDrivers)

    def __init__(self):
        super(SectionSkillDrivers, self).__init__()

    def __repr__(self):
        return '<%s for %r>' %(self.__class__.__name__, self.__parent__)
    """
    zope.component.adapts(ICourse)
    implements(interfaces.ISkillDrivers)

    def __repr__(self):
        return '<%s for %r>' %(self.__class__.__name__, self.__parent__)

    def __init__(self):
        super(SkillDrivers, self).__init__()

    def _getSkillDrivers(self):
        skillDrivers = list(super(SkillDrivers, self).values())
        return skillDrivers

    def keys(self):
        keys = []
        keys.extend([sd.__name__ for sd in self._getSkillDrivers()])
        return keys

    def __contains__(self, key):
        return key in self.keys()

    def items(self):
        return [(sd.__name__, sd)
                for sd in self._getSkillDrivers()]

    def __getitem__(self, key):
        for k, v in self.items():
            if k == key:
                return v
        return super(SkillDrivers, self).__getitem__(key)

    def values(self):
        return self._getSkillDrivers()

    def __len__(self):
        return len(self._getSkillDrivers())

    def __iter__(self):
        return iter(self.keys())

    def get(self, key, default=None):
        if key in self:
            return self[key]
        else:
            return default


class SectionSkillDriverProxy(LocationProxy):
    """Tags a skill driver with a section."""

    implements(interfaces.ISkillDriver)

    def __init__(self, ob, container=None, name=None):
        super(SectionSkillDriverProxy, self).__init__(ob, container, name)
        self.instructors = container.__parent__.instructors


class SectionSkillDrivers(btree.BTreeContainer):
    """See interfaces.ISectionSkillDrivers

    This gets a section by adapting a section.
    i.e. ISectionSkillDrivers(someSection) -> SectionSkillDrivers
    """

    zope.component.adapts(ISection)
    implements(interfaces.ISectionSkillDrivers)

    def __repr__(self):
        return '<%s for %r>' %(self.__class__.__name__, self.__parent__)

    def __init__(self):
        super(SectionSkillDrivers, self).__init__()

    def _getSkillDrivers(self):
        skillDrivers = list(super(SectionSkillDrivers, self).values())
        for course in self.__parent__.courses:
            sds = interfaces.ISkillDrivers(course).values()
            skillDrivers.extend([SectionSkillDriverProxy(sd, self,
                                                         '%s-%s' % (course.__name__,
                                                                    sd.__name__))
                                 for sd in sds])
        return skillDrivers

    def keys(self):
        keys = []
        keys.extend([sd.__name__ for sd in self._getSkillDrivers()])
        return keys

    def __contains__(self, key):
        return key in self.keys()

    def items(self):
        return [(sd.__name__, sd)
                for sd in self._getSkillDrivers()]

    def __getitem__(self, key):
        for k, v in self.items():
            if k == key:
                return v
        return super(SectionSkillDrivers, self).__getitem__(key)

    def values(self):
        return self._getSkillDrivers()

    def __len__(self):
        return len(self._getSkillDrivers())

    def __iter__(self):
        return iter(self.keys())

    def get(self, key, default=None):
        if key in self:
            return self[key]
        else:
            return default


class CourseLeaderCrowd(Crowd):
    """Crowd that contains only the leaders of courses."""

    def __init__(self, context):
        self.context = context

    def contains(self, principal):
        person = IPerson(principal, None)
        course = self.context.__parent__
        return person in course.leaders


# adapter from ICourse to ISkillDrivers via annotations.
getSkillDrivers = zope.annotation.factory(
    SkillDrivers, 'skilldrivers')


# adapter from ISection to ISectionSkillDrivers via annotations.
getSectionSkillDrivers = zope.annotation.factory(
    SectionSkillDrivers, 'skilldrivers')


class CanDoExternalActivity(object):

    implements(IExternalActivity)
    zope.component.adapts(interfaces.ISkillDriver)

    def __init__(self, context):
        self.skilldriver = context
        self.gradebook = ICompetencyGradebook(context)
        self.__parent__ = context
        self.source = ""
        self.external_activity_id = ""

    @property
    def title(self):
        return self.skilldriver.title

    @property
    def description(self):
        return self.skilldriver.description

    def getGrade(self, student):
        numComps = totalPoints = 0
        for competency in self.skilldriver.getCompetencies():
            ev = self.gradebook.getEvaluation(student, competency)
            if ev is None or ev.value == UNSCORED:
                continue
            numComps += 1
            totalPoints += ev.value.value
        numCompsTimes3 = numComps * 3
        if numComps:
            return Decimal(totalPoints) / Decimal(numCompsTimes3)
        return None

    def __eq__(self, other):
        return other is not None and \
               self.source == other.source and \
               self.external_activity_id == other.external_activity_id


class CanDoExternalActivities(object):

    implements(IExternalActivities)

    title = u"CanDo"
    source = "cando.external_activities"

    def __init__(self, context):
        self.context = context
        self.__parent__ = context

    def getExternalActivities(self):
        result = []
        intids = getUtility(IIntIds)
        section_skilldrivers = interfaces.ISectionSkillDrivers(self.context)
        for skilldriver in section_skilldrivers.values():
            external_activity = IExternalActivity(skilldriver)
            external_activity.source = self.source
            skilldriver_external_id = str(intids.getId(skilldriver))
            external_activity.external_activity_id = skilldriver_external_id
            result.append(external_activity)
        return result

    def getExternalActivity(self, external_activity_id):
        intids = getUtility(IIntIds)
        try:
            external_activity_id = int(external_activity_id)
        except (ValueError,):
            return None
        skilldriver = intids.queryObject(external_activity_id)
        if skilldriver is not None:
            return IExternalActivity(skilldriver)
