#
# CanDo - Web based competency tracking system.
# Copyright (C) 2006 Arlington Public Schools, Arlington, VA and The
# Shuttleworth Foundation.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""Virginia Competency Implementation

$Id$
"""
from socket import gethostname
import os.path

from persistent import Persistent
from persistent.list import PersistentList
from persistent.dict import PersistentDict

import zope.component
from zope.interface import implements, implementer
from zope.schema.fieldproperty import FieldProperty
from zope import annotation
from zope.traversing.api import getParents, getName
from zope.container.btree import BTreeContainer
from zope.intid import IntIds
from zope.intid.interfaces import IIntIds
from zope.catalog.catalog import Catalog
from zope.catalog.interfaces import ICatalog
from zope.catalog.field import FieldIndex

import schooltool.relationship
from schooltool.relationship.interfaces import IRelationshipLinks
from schooltool.relationship.interfaces import NoSuchRelationship
from schooltool.relationship import URIObject, RelationshipSchema
from schooltool.relationship import getRelatedObjects
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.requirement.requirement import Requirement
from schooltool.app.app import InitBase, getSchoolToolApplication
from schooltool.traverser.traverser import AdapterTraverserPlugin

from cando.virginia import interfaces

from zope.i18nmessageid import MessageFactory

_ = MessageFactory('cando')

GLOBAL_COMPETENCIES_REGISTRY_KEY = "cando.competency.global_registry"
COMPETENCY_INTIDS_KEY = 'cando.virginia.competency.intids'
COMPETENCY_CATALOG_KEY = 'cando.virginia.competency.catalog'
COMPETENCY_ID_UTILITY_KEY = 'cando.virginia.comeptency.idutility'


URICoreCompetencies = URIObject('http://schooltool.org/ns/core_comps',
     'CoreCompetencies', 'The core competencies.')
URIAdditionalComps = URIObject('http://schooltool.org/ns/addtional_comps',
    'AdditionalCompetencies', 'Any additional competencies.')
URICompetencyRef = URIObject('http://schooltool.org/ns/leadership/compref',
    'CompetencyRef', 'A role of a competency that is assigned.')
URICompetencyList = URIObject('http://schooltool.org/ns/leadership/comp_list',
    'CompetencyReferer', 'The object that has assigned competencies.')

CoreCompetencies = RelationshipSchema(URICoreCompetencies,
                                      compref=URICompetencyRef,
                                      target=URICompetencyList)
AdditionalComps =  RelationshipSchema(URIAdditionalComps,
                                      compref=URICompetencyRef,
                                      target=URICompetencyList)

class ScoreSystemRequirement(Requirement):
    '''An abstract class for things that want to inherit score systems.'''

    implements(interfaces.IScoreSystemRequirement)

    _scoresystem = None

    def _getScoreSystem(self):
        ss = None
        if self._scoresystem is None:
            if interfaces.IScoreSystemRequirement.providedBy(self.__parent__):
                ss= self.__parent__.scoresystem
            else:
                ss = zope.component.queryUtility(interfaces.IUniversalScoreSystem)
                if ss is None:
                    try:
                        ss = ISchoolToolApplication(None).get('scoresystem', None)
                    except ValueError:
                        pass #failed to adapt
        else:
            ss = self._scoresystem
        return ss

    def _setScoreSystem(self, s):
        self._scoresystem = s

    scoresystem = property(_getScoreSystem, _setScoreSystem)


class Competencies(Requirement):
    '''A container of competency references.'''
    implements(interfaces.ICompetencies,
               annotation.interfaces.IAttributeAnnotatable)


class Competency(Requirement):
    implements(interfaces.ICompetency)


class CompetencyReference(Requirement):
    implements(interfaces.ICompetencyReference,
               annotation.interfaces.IAttributeAnnotatable)

    SOLCorrelations = None
    modified = None
    deprecated = False

    def __init__(self, id, globalId, optionalFlag=u"required"):
        super(CompetencyReference, self).__init__(globalId)
        self.id = id
        self.globalId = globalId
        self.optionalFlag = optionalFlag
        self.SOLCorrelations = PersistentDict()
        self.modified = PersistentList()
        self.deprecated = False

    def setSOLCorrelations(self, correlations):
        self.SOLCorrelations = PersistentDict()
        for category, ids in correlations.items():
            self.SOLCorrelations[category] = PersistentList(ids)

    @property
    def competency(self):
        global_comps = getAppGlobalCompetencies(getSchoolToolApplication())
        return global_comps[self.globalId]

    def __repr__(self):
        return '<%s to %r>' % (self.__class__.__name__,
                               self.competency.title)


def assignCoreCompetencies(compref_tree_node, target):
    """Assign core competencies to a target course, section, etc."""
    assigned = interfaces.ICompetencyList(target)
    for compref in getFlatCompetencies(compref_tree_node):
        if compref not in assigned:
            CoreCompetencies(compref=compref, target=target)


def getCoreCompetencies(source):
    comps = []
    links = IRelationshipLinks(source)
    for link in links.getLinksByRole(URICompetencyRef):
        if link.rel_type == URICoreCompetencies:
            comps.append(link.target)
    return comps


def assignAdditionalCompetencies(compref_tree_node, target):
    """Assign additional competencies to a target course, section, etc."""
    assigned = interfaces.ICompetencyList(target)
    for compref in getFlatCompetencies(compref_tree_node):
        if compref not in assigned:
            AdditionalComps(compref=compref, target=target)


def getAdditionalCompetencies(source):
    comps = []
    links = IRelationshipLinks(source)
    for link in links.getLinksByRole(URICompetencyRef):
        if link.rel_type == URIAdditionalComps:
            comps.append(link.target)
    return comps


def unlink(relationship_link, target):
    """Helper extension to the schooltool.relationship library"""
    try:
        schooltool.relationship.unrelate(
            relationship_link.rel_type,
            (target, relationship_link.my_role),
            (relationship_link.target, relationship_link.role))
    except NoSuchRelationship:
        pass # it was a loop, so we tried to delete it twice


def unassignAllCoreCompetencies(target):
    links = list(IRelationshipLinks(target).getLinksByRole(URICompetencyRef))
    for link in links:
        if link.rel_type == URICoreCompetencies:
            unlink(link, target)


def unassignCompetencies(compref_tree_node, target):
    comps_to_unassign = getFlatCompetencies(compref_tree_node)
    for link in list(IRelationshipLinks(target)):
        if link.target in comps_to_unassign:
            unlink(link, target)


def getAppCompetencyRefTree(app):
    """Adapt an ``ISchoolToolApplication`` object to ``ICompetencies``."""
    return app['competencies']


def getAppGlobalCompetencies(app):
    """Get the competencies mapped by their global IDs"""
    return app[GLOBAL_COMPETENCIES_REGISTRY_KEY]


def getCompetencyFromCompRef(compRef):
    """Adapt an ``ICompetencyReference`` object to ``ICompetency``."""
    comps = getAppGlobalCompetencies(ISchoolToolApplication(None))
    return comps[compRef.globalId]


def getFlatCompetencies(rootComp):
    """This builds a flat list of a competency tree.

    It only includes competencies and not competency groups.
    """
    if interfaces.ICompetencyReference.providedBy(rootComp):
        competencies = [rootComp]
    else:
        competencies = []
        if hasattr(rootComp, 'values'):
            for comp in rootComp.values():
                competencies += getFlatCompetencies(comp)
        else:
            competencies += rootComp
    return competencies


def listAssignedCompetencies(obj):
    finalComps = []
    for comp in getRelatedObjects(obj, URICompetencyRef):
        if interfaces.ICompetencyReference.providedBy(comp):
            finalComps.append(comp)
    return sorted(finalComps, key=lambda c: c.id)


def listSectionCompetencies(section):
    competencies = listAssignedCompetencies(section)
    for course in section.courses:
        competencies.extend(listAssignedCompetencies(course))
    return sorted(list(set(competencies)), key=lambda c: c.id)


def buildCompetencyDictByCourse(obj):
    competencies = {}

    # only list competencies directly assigned to the object
    comp_list = listAssignedCompetencies(obj)

    for comp in comp_list:
        course = comp.__parent__.__parent__.title
        competencies[course] = {}

    # build the dictionary from the globalIds
    for comp in comp_list:
        group = comp.__parent__
        course = group.__parent__
        title = "%s - %s" % (course.title, group.title)
        course = comp.__parent__.__parent__.title
        if title not in competencies[course].keys():
            competencies[course][title] = []
        competencies[course][title].append(comp)

    # return the dictionary
    return competencies


def buildCompetencyDict(obj):
    """Build a dictionary whose keys are competency group titles with course
    names (i.e. `Animal Husbandry - Pet Care`) and whose values are competency
       lists for the group."""

    competencies = {}
    for comp in interfaces.ICompetencyList(obj):
        group = comp.__parent__
        course = group.__parent__
        title = "%s - %s" % (course.title, group.title)
        if title not in competencies.keys():
            competencies[title] = []
        competencies[title].append(comp)

    return competencies


class GlobalIdUtility(Persistent):
    '''See .interfaces.IGlobalIdUtility'''

    implements(interfaces.IGlobalIdUtility)

    prefix = FieldProperty(interfaces.IGlobalIdUtility['prefix'])

    def __init__(self):
        self.prefix = 'http://competencies.%s/' % gethostname()
        self._count = 1

    def _formatName(self, name):
        name = name.strip()
        newname = ''
        for char in name:
            if char.isalpha():
                newname += char.lower()
            elif char.isspace():
                newname += '-'
        return newname

    def _checkName(self, name):
        comps = getAppGlobalCompetencies(ISchoolToolApplication(None))
        return name not in comps

    def generate(self, name=None, object=None):
        '''See .interfaces.IGlobalIdUtility'''
        if name is not None:
            name = self._formatName(name)
        else:
            if object is not None:
                name = self._formatName(getName(object))
            else:
                name = str(self._count)
                self._count += 1
        if object is not None:
            parents = [self._formatName(getName(obj)) for obj in reversed(getParents(object))]
            name = os.path.join(os.path.join(*parents), name)
        name = os.path.join(self.prefix, name)

        while not self._checkName(name):
            split = name.rsplit('-', 1)
            if len(split) != 1:
                name, count = split
            else:
                count = split[-1]
            try:
                count = int(count) + 1
            except ValueError:
                count = 2
            name = "%s-%s" % (name, count)
        return name


@implementer(interfaces.IGlobalIdUtility)
@zope.component.adapter(ISchoolToolApplication)
def getGlobalIdUtility(app):
    sm = app.getSiteManager()
    return sm.queryUtility(interfaces.IGlobalIdUtility)


GlobalIdTraverserPlugin = AdapterTraverserPlugin('GlobalCompetencyIdUtility',
                                                 interfaces.IGlobalIdUtility)


class CompetencyInit(InitBase):

    def __call__(self):
        # Competency reference tree
        # XXX: the key probably should be renamed to something more sane
        self.app['competencies'] = Competencies(_('Competency reference tree'))

        # The registry of actual competencies, by their global ID
        self.app[GLOBAL_COMPETENCIES_REGISTRY_KEY] = BTreeContainer()

        sm = self.app.getSiteManager()

        # Set up Integer Id Utility if one isn't already set up.
        intids = IntIds()
        sm[COMPETENCY_INTIDS_KEY] = intids
        if sm.queryUtility(IIntIds) is None:
            sm.registerUtility(intids, IIntIds)

        # Set up competency catalog
        catalog = Catalog()
        sm[COMPETENCY_CATALOG_KEY] = catalog
        sm.registerUtility(catalog, ICatalog, COMPETENCY_CATALOG_KEY)

        # Set up global id index for competency catalog
        globalIdIndex = FieldIndex(
            interface=interfaces.ICompetencyReference,
            field_name='globalId'
            )
        catalog[u'globalId'] = globalIdIndex

        #setup the global id utility
        idutil = GlobalIdUtility()
        sm[COMPETENCY_ID_UTILITY_KEY] = idutil
        sm.registerUtility(idutil, interfaces.IGlobalIdUtility)

